"""
SeaLion is a simple machine learning and data science library.

=============================================================

Designed with beginners in mind, it has rich documentation via Pydoc and algorithms that span from the most basic
to more modern approaches. It is meant to help beginners navigate it all, and the documentation not only explains t
he models and their respective functions but also what they are and when to use them. Emphasis was also put on creating
new functions to make it interesting for those who are just getting started and seasoned ml-engineers alike.

I hope you enjoy it!
- Anish Lakkapragada 2021
"""

from . import regression  # passed
from . import decision_trees
from . import DimensionalityReduction
from . import ensemble_learning
from . import naive_bayes
from . import nearest_neighbors
from . import unsupervised_clustering
from . import utils
from . import neural_networks


import time
import os
import subprocess
import pickle
import shutil
import sys

start = time.time()

org_dir = os.getcwd()
dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)

VERSION_NUMBER = "4.3.9"
PYTHON_RUNNING_VERSION = str(sys.version_info.major) + "." + str(sys.version_info.minor)


def read_pickle_file(file) :
    with open(file, 'rb') as f :
        stuff = pickle.load(f)
    return stuff

if os.path.exists("cython_ran.pickle") :
    if read_pickle_file("cython_ran.pickle") == VERSION_NUMBER :
        # then we are up to date here
        pass
    else :
        # this means that it was a different version
        var = subprocess.Popen(
            f"python{PYTHON_RUNNING_VERSION} setup.py build_ext --inplace",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        while True:
            # files are currently being generated
            try:
                from . import cython_decision_tree_functions
                from . import cython_knn
                from . import cython_naive_bayes
                from . import cython_unsupervised_clustering
                from . import cython_tsne
                from . import cython_ensemble_learning
                from . import cython_mixtures
                break
            except Exception:
                if time.time() - start > 500:
                    print(
                        "Cython compilation files unable to load. Please raise this error on github, "
                        "and especially include your architecture, OS, and Python version. Thank you!"
                    )

                    break

        with open("cython_ran.pickle", "wb") as f:
            pickle.dump(VERSION_NUMBER, f) # save the version number
else:
    var = subprocess.Popen(
        f"python{PYTHON_RUNNING_VERSION} setup.py build_ext --inplace",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    while True:
        # files are currently being generated
        try:
            from . import cython_decision_tree_functions
            from . import cython_knn
            from . import cython_naive_bayes
            from . import cython_unsupervised_clustering
            from . import cython_tsne
            from . import cython_ensemble_learning
            from . import cython_mixtures
            break
        except Exception:
            if time.time() - start > 500:
                print(
                    "Cython compilation files unable to load. Please raise this error on github, "
                    "and especially include your architecture, OS, and Python version. Thank you!"
                )

                break

    with open("cython_ran.pickle", "wb") as f:
        pickle.dump(VERSION_NUMBER, f)


os.chdir(org_dir)



"""
@author : Anish Lakkapragada
@date : 1 - 23 - 2021

The loss functions are really simple. You just need to understand whether it is a classification or regression task.
All losses will be set in the model.finalize() model.
"""
import numpy as np
import warnings
from scipy.special import softmax as sfmx_indiv

warnings.filterwarnings("ignore", category=RuntimeWarning)


class Loss:
    def __init__(self):
        self.SGD = False

    def loss(self, y, y_pred):
        pass

    def grad(self, y, y_pred):
        pass


class MSE(Loss):
    """
    MSE stands for mean-squared error, and its the loss you'll want to use for regression. To set it in the model.finalize()
    method just do :
    >>> from sealion import neural_networks as nn
    >>> model = nn.models.NeuralNetwork(layers_list)
    >>> model.finalize(loss = nn.loss.MSE(), optimizer = ...)

    and you're all set!

    """

    def __init__(self):
        super().__init__()
        self.type_regression = True

    def loss(self, y, y_pred):
        error = np.sum(np.power(y_pred - y, 2)) / (2 * len(y))
        return error

    def grad(self, y, y_pred):
        return (y_pred - y) / len(y)


def softmax(x):
    softmax_output = np.apply_along_axis(sfmx_indiv, 1, x)
    return softmax_output


class CrossEntropy(Loss):
    """
    This loss function is for classification problems. I know there's a binary log loss and then a multi-category cross entropy
    loss function for classification, but they're essentially the same thing so I thought using one class would make it easier.
    Remember to use one-hot encoded data for this to work (check out utils.)

    If you are using this loss function, make sure your last layer is Softmax and vice versa. Otherwise, annoying error
    messages will occur.

    To set this in the model.finalize() method do :
    >>> from sealion import neural_networks as nn
    >>> model = nn.models.NeuralNetwork()
    ... add the layers ...
    >>> model.add(nn.layers.Softmax()) #last layer has to be softmax
    >>> model.finalize(loss = nn.loss.CrossEntropy(), optimizer = ...)

    and that's all there is to it.
    """

    def __init__(self):
        super().__init__()
        self.type_regression = False

    def loss(self, y, y_pred):
        return np.sum(y * np.log(y_pred + 1e-20)) / len(
            y
        )  # now give the crossentropy loss

    def grad(self, y, y_pred):
        y_pred = softmax(y_pred)
        return (y_pred - y) / len(y)  # give the sexy partial derivative




from Cython.Build import cythonize
from setuptools import setup, Extension
import numpy
import warnings

warnings.filterwarnings("ignore")

files = [
    "cython_decision_tree_functions",
    "cython_tsne",
    "cython_unsupervised_clustering",
    "cython_naive_bayes",
    "cython_knn",
    "cython_ensemble_learning",
    "cython_mixtures"
]


def setup_indiv(file_name):
    extension = Extension(
        file_name, [file_name + ".pyx"], include_dirs=[numpy.get_include()]
    )
    setup(ext_modules=cythonize(extension))


from joblib import parallel_backend, delayed, Parallel

for file_name in files :
    setup_indiv(file_name)




"""
@author : Anish Lakkapragada
@date : 1 - 23 - 2021

The classes that glue the layers, losses, and optimizers all together. Here you will find the neural network class for
you to build deep neural networks and we also have a map-reduce implementation of neural networks for fun. Highly
recommend you see the examples on GitHub as they will help you more than anything else.
"""

import numpy as np
import warnings
from joblib import Parallel, delayed
from tqdm import tqdm
from multiprocessing import cpu_count
import random


def r2_score(y_pred, y_test):
    num = np.sum(np.power(y_test - y_pred, 2))
    denum = np.sum(np.power(y_test - np.mean(y_test), 2))
    return 1 - num / denum


def _perc_correct(y_pred, y_test):
    return np.sum((y_pred == y_test).astype("int")) / len(y_pred)


class NeuralNetwork:
    """
    This class is very rich and packed with methods, so I figured it would be best to have this tutorial guide you
    on a tutorial on the "hello world" of machine learning.

    There are two ways to initialize this model with its layers, through the init() or through the .add() function.

    The first way :
    >>> import neural_networks as nn
    >>> layers = [nn.layers.Flatten(), nn.layers.Dense(784, 64, activation = nn.layers.ReLU()), nn.layers.Dense(64, 32, activation = nn.layers.ReLU()),
    >>> nn.layers.Dense(32, 10, activation = nn.layers.Softmax())]
    >>> model = nn.models.NeuralNetwork(layers)

    Or you can go through the second way :
    >>> import neural_networks as nn
    >>> model = nn.models.NeuralNetwork()
    >>> model.add(nn.layers.Flatten())
    >>> model.add(nn.layers.Dense(784, 64, activation = nn.layers.ReLU()))
    >>> model.add(nn.layers.Dense(64, 32, activation = nn.layers.ReLU()))
    >>> model.add(nn.layers.Dense(32, 10, activation = nn.layers.ReLU()))

    Either way works just fine.

    Next, you will want to perhaps see how complex the model is (having too many parameters means a model can easily
    overfit), so for that you can just do :
    >>> num_params = model.num_parameters() #returns an integer
    >>> assert num_params == 52650

    Looks like our model will be pretty complex. Next up is finalizing and training.
    >>> model.finalize(loss = nn.loss.CrossEntropy(), optimizer = nn.optimizers.Adam())

    Here we use cross-entropy loss for this classification problem and the Adam optimizer.

    Onto training. Assuming our data is 60k * 28 * 28 (sounds a lot like MNIST) in the variable X_train and we have y_train is
    a one-hot encoded matrix size 60k * 10 (10 classes) we can do :

    >>> model.train(X_train, y_train, epochs = 20) #train for 20 epochs

    which will work fine. Here we have a batch_size of 32 (default), and the way we make this run fast is by making
    batch_size datasets, and training them in parallel via multithreading.

    If you want to change batch_size to 18 you could do:
    >>> model.train(X_train, y_train, epochs = 20, batch_size=18) #train for 20 epochs, batch_size 18

    If you want the gradients to be calculated over the entire dataset (this will be much longer) you can do :
    >>> model.full_batch_train(X_train, y_train, epochs = 20)

    Lastly there's also something known as mini-batch gradient descent, which just does gradient descent but randomly chooses
    a percent of the dataset for gradients to be calculated upon. This cannot be parallelized, but still runs fast :
    >>> model.mini_batch_train(X_train, y_train, N = 0.1) #here we take 10% of X_train or 6000 data-points randomly selected for calculating gradients at a time

    All of these methods have a show_loop parameter on whether you want to see the tqdm loop, which is set true by default.

    Now that we have trained our model, time to test and use it.
    To evaluate it, given X_test (shape 10k * 28 * 28) and y_test (shape 10k * 10), we can feed this into
    our evaluate() function :

    >>> model.evaluate(X_test, y_test)

    This gives us the loss, which can be not so interpretable - so we have some other options.

    If you are doing a classification problem as we are here you may do instead :
    >>> model.categorical_evaluate(X_test, y_test)

    which just gives what percent of X_test was classified correctly.

    For regression just do :
    >>> model.regression_evaluate(X_test, y_test)

    which gives the r^2 value of the predictions on X_test. If you are doing this make sure that y_test is not one-hot encoded.

    To predict on data we can just do :
    >>> predictions = model.predict(X_test)

    and if we want this in reverted one-hot-encoded form all we need to do is this :
    >>> from utils import revert_softmax
    >>> predictions = revert_softmax(predictions)

    Storing around 53,000 parameters in matrices isn't much, but for good practice let's store it.
    Using the give_parameters() method we can save our weights in a list :
    >>> parameters = model.give_parameters()

    Then we may want to pickle this using the given method, into a file called "MNIST_pickled_params" :
    >>> file_name = "MNIST_pickled_params"
    >>> model.pickle_params(FILE_NAME=file_name)

    Now this is the beauty - let's say a few weeks from now we come back and realize we probably should train for 20 epochs.
    BUT - we don't want to have to restart training (imagine this was a really big network.)

    So we can just load the weights using the pickle module, build the same model architecture, insert the params, and hit train()!!

    >>> import pickle
    >>> with open('MNIST_pickled_params.pickle', 'rb') as f :
    >>>  params = pickle.load(f)

    Now we can create the model structure (must be the EXACT same) :
    >>> model = nn.models.NeuralNetwork()
    >>> model.add(nn.layers.Flatten())
    >>> model.add(nn.layers.Dense(784, 64, activation = nn.layers.ReLU()))
    >>> model.add(nn.layers.Dense(64, 32, activation = nn.layers.ReLU()))
    >>> model.add(nn.layers.Dense(32, 10, activation = nn.layers.ReLU()))

    Obviously we want to finalize our model for good practice :
    >>> model.finalize(loss = nn.loss.CrossEntropy(), optimizer = nn.optimizers.Adam())

    and we can enter in the weights & biases using the enter_parameters() :
    >>> model.enter_parameters(params) #must be params given from the give_parameters() method

    and train! :
    >>> model.train(X_train, y_train, epochs = 100) #let's try 100 epochs this time

    The reason this is such a big deal is because we don't have to start training from scratch all over again. The model
    will not have to start from 0% accuracy, but may start with 80% given the params belonging to our partially-trained model
    we loaded from the pickle file. This is of course not a big deal for something like MNIST but will be for bigger model architectures, or datasets.

    Of course there's a lot more things we could've changed, but I think that's pretty good for now!
    """

    def __init__(self, layers=None):
        from .layers import Dense, Flatten, Softmax, Dropout
        from .loss import CrossEntropy, softmax
        from .optimizers import Adam, SGD
        from ._utils_nn import revert_one_hot_nn, revert_softmax_nn

        self.Dense = Dense
        self.Flatten = Flatten
        self.Softmax = Softmax
        self.Dropout = Dropout
        self.CrossEntropy = CrossEntropy
        self.softmax = softmax
        self.Adam = Adam
        self.SGD = SGD
        self.revert_one_hot = revert_one_hot_nn
        self.revert_softmax = revert_softmax_nn

        self.layers = []
        if layers:  # self.layers = layers
            for layer in layers:
                self.layers.append(layer)
                if layer.activation:
                    self.layers.append(layer.activation)
        self.finalized = False
        self.sgd_indices = ...
        self.adam_t = 0

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        warnings.filterwarnings("ignore", category=np.ComplexWarning)

    def num_parameters(self):
        num_params = 0
        for layer in self.layers:
            if isinstance(layer, self.Dense):
                num_params += (
                    np.product(layer.parameters["weights"].shape)
                    + layer.parameters["bias"].shape
                )
        return num_params

    def enter_parameters(self, parameters):
        for layer_num in range(len(self.layers)):
            try:
                if (
                    parameters[layer_num]["weights"] == None
                    and parameters[layer_num]["bias"] == None
                ):
                    continue
            except Exception:
                self.layers[layer_num].parameters["weights"] = parameters[layer_num][
                    "weights"
                ]
                self.layers[layer_num].parameters["bias"] = parameters[layer_num][
                    "bias"
                ]

    def give_parameters(self):
        parameters = []
        for layer in self.layers:
            try:
                parameters.append(
                    {
                        "weights": layer.parameters["weights"],
                        "bias": layer.parameters["bias"],
                    }
                )
            except Exception:
                parameters.append({"weights": None, "bias": None})
        return parameters

    def pickle_params(self, FILE_NAME="NeuralNetwork_learnt_parameters"):
        parameters = NeuralNetwork.give_parameters(self)
        import pickle

        with open(FILE_NAME + ".pickle", "wb") as f:
            pickle.dump(parameters, f, protocol=pickle.HIGHEST_PROTOCOL)

    def add(self, layer):
        self.layers.append(layer)
        if layer.activation:
            self.layers.append(layer.activation)

    def forward(self, inputs):
        for layer in self.layers:
            if isinstance(layer, self.Softmax):
                continue
            inputs = layer.forward(inputs)
        return inputs

    def backward(self, grad):
        for layer in reversed(self.layers):
            if isinstance(layer, self.Softmax):
                continue
            grad = layer.backward(grad)
        return grad

    def finalize(self, loss, optimizer):
        """Both of these have to be the classes for the loss and optimizations"""
        self.loss, self.optimizer = loss, optimizer
        self.finalized = True

    def _sgd_updating_indices(self):
        """only call if the loss is SGD"""

        for layer_num in range(len(self.layers)):
            layer = self.layers[layer_num]
            layer.indices = self.sgd_indices
            self.layers[layer_num] = layer

    def _chunk_train(self, data_chunk, label_chunk):
        """the actual training code!"""
        predicted = NeuralNetwork.forward(self, data_chunk)
        for layer_num in range(len(self.layers)):
            if self.layers[layer_num].nesterov:
                """nesterov accel"""
                try:
                    self.layers[layer_num].parameters["weights"] += (
                        self.optimizer.momentum
                        * self.optimizer.momentum_params["weights" + str(layer_num)]
                    )
                    self.layers[layer_num].parameters["bias"] += (
                        self.optimizer.momentum
                        * self.optimizer.momentum_params["bias" + str(layer_num)]
                    )
                except Exception:
                    pass

        if isinstance(self.optimizer, self.Adam):
            self.adam_t += 1
            self.optimizer.t = self.adam_t

        if isinstance(self.optimizer, self.SGD):
            """select indices for the gradients and update all layers"""
            start_index = np.random.randint(len(predicted) - 3)
            self.sgd_indices = slice(start_index, len(predicted) - 1)
            NeuralNetwork._sgd_updating_indices(self)

        grad = self.loss.grad(
            label_chunk[self.sgd_indices], predicted[self.sgd_indices]
        )  # calculate the dL/dY
        NeuralNetwork.backward(self, grad)
        self.optimizer.update(self.layers)

    def train(self, x_train, y_train, epochs=1, batch_size=32, show_loop=True):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        if isinstance(self.layers[-1], self.Softmax) and not isinstance(
            self.loss, self.CrossEntropy
        ):
            raise ValueError(
                "If the last layer activation is softmax, you must be using CrossEntropy loss."
            )

        if isinstance(self.loss, self.CrossEntropy) and not isinstance(
            self.layers[-1], self.Softmax
        ):
            raise ValueError(
                "If the loss function is crossentropy, you must be using Softmax as your last layer."
            )

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(y_train.shape) != 2:
            raise ValueError("y_train must be 2D.")
        if not isinstance(self.layers[0], self.Flatten) and len(x_train.shape) > 2:
            raise ValueError(
                "There must be a Flatten layer in the beginning if you are working with data that is not "
                "a matrix (e.g. something that is 60k * 28 * 28 not 60k * 784)"
            )

        # perform batch operations
        if batch_size > x_train.shape[0]:
            warnings.warn(
                "Batch size is more than the number of samples, so we have set batch_size = 1 and are just performing full batch gradient descent."
            )
            batch_size = 1

        x_train = np.array_split(x_train, batch_size)
        y_train = np.array_split(y_train, batch_size)
        self.optimizer.setup(self.layers)

        epochs = (
            tqdm(range(epochs), position=0, ncols=100) if show_loop else range(epochs)
        )
        evaluation_batch = random.randint(0, len(x_train) - 1)
        evaluation_method = (
            NeuralNetwork.regression_evaluate
            if self.loss.type_regression
            else NeuralNetwork.categorical_evaluate
        )

        for epoch in epochs:
            Parallel(prefer="threads")(
                delayed(NeuralNetwork._chunk_train)(self, data_chunk, label_chunk)
                for data_chunk, label_chunk in zip(x_train, y_train)
            )

            if show_loop:
                epochs.set_description(
                    "Acc : "
                    + str(
                        round(
                            evaluation_method(
                                self,
                                x_train[evaluation_batch],
                                y_train[evaluation_batch],
                            )
                            * 100,
                            2,
                        )
                    )
                    + "%"
                )

        # adjust dropout
        for layer_num in range(len(self.layers)):
            """adjust the weights and biases if there is dropout"""
            try:
                layer = self.layers[layer_num]
                if isinstance(layer, self.Dropout):
                    receiving_layer = self.layers[
                        layer_num + 1
                    ]  # this is the layer that receives the dropout
                    receiving_layer.parameters["weights"] *= 1 - layer.dr
                    receiving_layer.parameters["bias"] *= 1 - layer.dr
                    self.layers[layer_num + 1] = receiving_layer
            except Exception:
                pass

    def full_batch_train(self, x_train, y_train, epochs=1, show_loop=True):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        if isinstance(self.layers[-1], self.Softmax) and not isinstance(
            self.loss, self.CrossEntropy
        ):
            raise ValueError(
                "If the last layer activation is softmax, you must be using CrossEntropy loss."
            )

        if isinstance(self.loss, self.CrossEntropy) and not isinstance(
            self.layers[-1], self.Softmax
        ):
            raise ValueError(
                "If the loss function is crossentropy, you must be using Softmax as your last layer."
            )

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(y_train.shape) != 2:
            raise ValueError("y_train must be 2D.")
        if not isinstance(self.layers[0], self.Flatten) and len(x_train.shape) > 2:
            raise ValueError(
                "There must be a Flatten layer in the beginning if you are working with data that is not "
                "a matrix (e.g. something that is 60k * 28 * 28 not 60k * 784)"
            )

        epochs = (
            tqdm(range(epochs), position=0, ncols=100) if show_loop else range(epochs)
        )
        evaluation_method = (
            NeuralNetwork.regression_evaluate
            if self.loss.type_regression
            else NeuralNetwork.categorical_evaluate
        )
        self.optimizer.setup(self.layers)  # setup the optimizer

        for epoch in epochs:
            NeuralNetwork._chunk_train(self, x_train, y_train)

            if show_loop == True:
                epochs.set_description(
                    "Acc : "
                    + str(round(evaluation_method(self, x_train, y_train) * 100, 2))
                    + "%"
                )

        # adjust dropout
        for layer_num in range(len(self.layers)):
            """adjust the weights and biases if there is dropout"""
            try:
                layer = self.layers[layer_num]
                if isinstance(layer, self.Dropout):
                    receiving_layer = self.layers[
                        layer_num + 1
                    ]  # this is the layer that receives the dropout
                    receiving_layer.parameters["weights"] *= 1 - layer.dr
                    receiving_layer.parameters["bias"] *= 1 - layer.dr
                    self.layers[layer_num + 1] = receiving_layer
            except Exception:
                pass

    def mini_batch_train(self, x_train, y_train, epochs=1, N=0.2, show_loop=True):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        if isinstance(self.layers[-1], self.Softmax) and not isinstance(
            self.loss, self.CrossEntropy
        ):
            raise ValueError(
                "If the last layer activation is softmax, you must be using CrossEntropy loss."
            )

        if isinstance(self.loss, self.CrossEntropy) and not isinstance(
            self.layers[-1], self.Softmax
        ):
            raise ValueError(
                "If the loss function is crossentropy, you must be using Softmax as your last layer."
            )

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(y_train.shape) != 2:
            raise ValueError("y_train must be 2D.")
        if not isinstance(self.layers[0], self.Flatten) and len(x_train.shape) > 2:
            raise ValueError(
                "There must be a Flatten layer in the beginning if you are working with data that is not "
                "a matrix (e.g. something that is 60k * 28 * 28 not 60k * 784)"
            )

        if N == 1:
            raise ValueError("N cannot be equal to 1, has to be between 0 and 1.")

        N *= len(x_train)
        N = round(N)

        epochs = (
            tqdm(range(epochs), position=0, ncols=100) if show_loop else range(epochs)
        )
        evaluation_method = (
            NeuralNetwork.regression_evaluate
            if self.loss.type_regression
            else NeuralNetwork.categorical_evaluate
        )
        self.optimizer.setup(self.layers)

        for epoch in epochs:

            # set up the mini-batch
            start_index = random.randint(0, len(x_train) - N - 1)
            end_index = start_index + N
            data_chunk, label_chunk = (
                x_train[start_index:end_index],
                y_train[start_index:end_index],
            )

            NeuralNetwork._chunk_train(self, data_chunk, label_chunk)

            if show_loop == True:
                epochs.set_description(
                    "Acc : "
                    + str(round(evaluation_method(self, x_train, y_train) * 100, 2))
                    + "%"
                )

        # adjust dropout
        for layer_num in range(len(self.layers)):
            """adjust the weights and biases if there is dropout"""
            try:
                layer = self.layers[layer_num]
                if isinstance(layer, self.Dropout):
                    receiving_layer = self.layers[
                        layer_num + 1
                    ]  # this is the layer that receives the dropout
                    receiving_layer.parameters["weights"] *= 1 - layer.dr
                    receiving_layer.parameters["bias"] *= 1 - layer.dr
                    self.layers[layer_num + 1] = receiving_layer
            except Exception:
                pass

    def predict(self, x_test):

        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        x_test = np.array(x_test)
        y_pred = NeuralNetwork.forward(self, x_test)
        if isinstance(self.layers[-1], self.Softmax):
            y_pred = self.softmax(y_pred)

        if self.loss.type_regression:
            return y_pred
        else:
            return np.round_(y_pred)

    def evaluate(self, x_test, y_test):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")

        y_pred = NeuralNetwork.predict(self, x_test)

        return self.loss.loss(y_test, y_pred)

    def regression_evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")

        y_pred = NeuralNetwork.predict(self, x_test)

        return np.mean(
            [
                r2_score(y_pred[:, col].flatten(), y_test[:, col].flatten())
                for col in range(y_pred.shape[1])
            ]
        )  # mean of each columns r^2

    def categorical_evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")

        y_pred = NeuralNetwork.predict(self, x_test)

        return _perc_correct(
            self.revert_softmax(y_pred).flatten(), self.revert_one_hot(y_test).flatten()
        )


class NeuralNetwork_MapReduce:
    """
    Map Reduce is a key algorithm to training neural networks these days. Maybe not as complicated as some other strategies,
    but as usual I couldn't resist the urge to try to see how this could work.

    Map Reduce is used to train datasets a lot faster - utilizing the multiple cores your computer might have. What it does
    is take a dataset and chop into however many cores you want to use (default is all of them), and then have each core train
    a model given that data. So if you use 16 cores, you will essentially have 64 models built on datasets that each are 1/16th
    the size of the given data.

    Then at test time, predictions will be ran through each model and will be averaged/rounded depending on whether its a
    regression or classification task. All the other methods of add(), finalize(), are mostly the same as with the
    NeuralNetwork() class.

    Let's say we are working with MNIST in variables X_train, and y_train. Well to first declare and finalize our model
    we could do :

    >>> import neural_networks as nn
    >>> model = nn.models.NeuralNetwork_MapReduce()
    >>> model.add(nn.layers.Flatten())
    >>> model.add(nn.layers.Dense(784, 64, activation = nn.layers.ReLU()))
    >>> model.add(nn.layers.Dense(64, 32, activation = nn.layers.ReLU()))
    >>> model.add(nn.layers.Dense(32, 10, activation = nn.layers.ReLU()))
    >>> model.finalize(loss = nn.loss.CrossEntropy(), optimizer = nn.optimizers.Adam())

    to train() it we could then do :
    >>> if model.train(X_train, y_train, epochs=20, num_cores=-1)

    which will work. I feel like I need to explain what num_cores is. It is simply the number of cores you want
    to be utilized. By default it is -1 which means all available cores. -2 would mean using all but 1, and -3 would
    mean using all but 2, -4 would mean using all but 3, etc. You could also set it to a positive number like 16, 100 -
    which is just the amount of cores utilized by the model. Another note is that this class doesn't have the show_loop
    argument.

    to predict() and evaluate() you could just do :
    >>> pred = model.predict(X_test)
    >>> categorical_evaluation = model.categorical_evaluate(X_test, y_test)

    to save the weights in a pickle file you could do :
    >>> file_name = "MNIST_MapReduce"
    >>> model.pickle_best_parameters(X_test, y_test, WEIGHT_FILE_NAME=file_name)

    Here what is going on is that the parameters you are getting pickled are the params of the best model.
    This makes MapReduce also useful as often times random initialization of parameters can effect the performance
    of the model, so running it a couple of times on many cores may help you find a good weight init.

    There's another method of give_best_parameters() which is also very similar to the pickle_best_parameters().

    This class was created to see what could happen, and I'm adding it here because hopefully you will find use in training
    many models to find the best model or using this to utilize all of your CPU cores. If you find a bug or an error,
    as usual please notify on Github or email me at anish.lakkapragada@gmail.com

    """

    def __init__(self, layers=None):
        from ._utils_nn import revert_one_hot_nn, revert_softmax_nn

        self.revert_softmax = revert_softmax_nn
        self.revert_one_hot = revert_one_hot_nn
        self.layers = []

        if layers:  # self.layers = layers
            for layer in layers:
                self.layers.append(layer)
                if layer.activation:
                    self.layers.append(layer.activation)
        self.finalized = False

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        warnings.filterwarnings("ignore", category=np.ComplexWarning)

    def add(self, layer):
        self.layers.append(layer)
        if layer.activation:
            self.layers.append(layer.activation)

    def forward(self, inputs):
        for layer in self.layers:
            inputs = layer.forward(inputs)
        return inputs

    def backward(self, grad):
        for layer in reversed(self.layers):
            grad = layer.backward(grad)
        return grad

    def finalize(self, loss, optimizer):
        """Both of these have to be the classes for the loss and optimizations"""
        self.loss, self.optimizer = loss, optimizer
        self.finalized = True

    def num_parameters(self):
        return self.models[0].num_parameters()

    def _train_individual_model(self, model_num):
        """here we train the individual model for the MapReduce algo"""
        model = self.models[model_num]
        data_chunk, label_chunk = self.x_train[model_num], self.y_train[model_num]
        model.full_batch_train(
            data_chunk, label_chunk, epochs=self.epochs, show_loop=False
        )
        self.models[model_num] = model  # update it to the train algorithm

    def give_best_parameters(self, x_test, y_test):
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")
        models = {}  # evaluation_score : class
        for model in self.models:
            score = model.evaluate(x_test, y_test)
            models[score] = model
        best_model = models[min(models)]
        return best_model.give_parameters()

    def pickle_best_parameters(
        self, x_test, y_test, FILE_NAME="NeuralNetwork_learnt_parameters"
    ):
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")
        best_params = NeuralNetwork_MapReduce.give_best_parameters(self, x_test, y_test)
        
        import pickle

        with open(FILE_NAME + ".pickle", "wb") as f:
            pickle.dump(best_params, f, protocol=pickle.HIGHEST_PROTOCOL)

    def train(self, x_train, y_train, epochs=1, batch_size=32, num_cores=-1):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        # store these for the above method
        self.epochs = epochs
        self.batch_size = batch_size

        x_train, y_train = np.array(x_train), np.array(y_train)

        if len(y_train.shape) != 2:
            raise ValueError("y_train must be 2D.")
        if not num_cores:
            raise ValueError("num_cores cannot be 0!")
        num_cores_used = cpu_count() + 1 + num_cores if num_cores < 0 else num_cores
        self.num_cores_used = num_cores_used

        # split the x_train and y_train into the number of cores

        x_train = np.array_split(x_train, num_cores_used)
        y_train = np.array_split(y_train, num_cores_used)
        self.x_train, self.y_train = x_train, y_train

        # train using each dataset

        self.models = []

        for core in range(num_cores_used):
            model = NeuralNetwork(self.layers)
            model.finalize(self.loss, self.optimizer)
            self.models.append(model)

        self.model_nums = range(len(self.models))
        Parallel()(
            delayed(NeuralNetwork_MapReduce._train_individual_model)(self, model_num)
            for model_num in self.model_nums
        )

        self.num_cores_used = num_cores_used  # save this for the predict method

    def predict(self, x_test):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        x_test = np.array(x_test)
        predictions_across_models = np.array(
            [model.predict(x_test) for model in self.models]
        )
        if self.loss.type_regression:  # todo fix this
            return np.sum(predictions_across_models, axis=0) / self.num_cores_used
        else:
            return np.round_(
                np.sum(predictions_across_models, axis=0) / self.num_cores_used
            )

    def evaluate(self, x_test, y_test):
        if not self.finalized:
            raise ValueError(
                "This model isn't finalized. Call it through model.finalize()."
            )

        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")

        y_pred = NeuralNetwork_MapReduce.predict(self, x_test)

        return self.loss.loss(y_test, y_pred)

    def regression_evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        # if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")

        y_pred = NeuralNetwork_MapReduce.predict(self, x_test)

        return np.mean(
            [
                r2_score(y_pred[:, col].flatten(), y_test[:, col].flatten())
                for col in range(y_pred.shape[1])
            ]
        )  # mean of each columns r^2

    def categorical_evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(y_test.shape) != 2:
            raise ValueError("y_test must be 2D.")

        y_pred = np.round_(NeuralNetwork_MapReduce.predict(self, x_test))

        return _perc_correct(self.revert_softmax(y_pred), self.revert_one_hot(y_test))




import numpy as np

class MultinomialNaiveBayes() :
    """
    Multinomial Naive Bayes is the most popular form of Naive Bayes there is. (Multinomial) Naive Bayes is well known for
    its usage in spam classifiers, and this specific module is just for that. You give your data,
    tokenized into numeric form (just means each word has a specific index), and you will get a
    spam classifier! You can then put your data in and get a prediction. Works for 2 or more categories/unique labels.
    Pretty fast with Cython.


    For example if your data was :

    [['get free money!'], ['buy 2 get 3'], ['Towards Data Science - New Article!']]

    And you tokenized it to :

    >>> data = [[1, 2, 3], [4, 5, 1, 6], [7, 8,9, 10, 11]]

    Your labels maybe (based on how "spammy" they are) :

    >>> labels = [1, 1, 0] # where 1 is spam and 0 is not

    You can feed this directly into this class like :
    >>> mnb = MultinomialNaiveBayes()
    >>> mnb.fit(data, labels)
    >>> mnb.predict() ...

    ----
    Methods
    fit(x_train, y_train) :
        ->> performs gradient descent
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)
    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> returns the predictions (0s and 1s) in a numpy array (1D)
    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> will the percent of predictions on x_test it got correct
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """

    def __init__(self):
        """Init, no args."""
        from .cython_naive_bayes import cy_MultinomialNaiveBayes
        self.inner_cython_mnb = cy_MultinomialNaiveBayes()
    def fit(self, x_train, y_train):
        """
        :param x_train: data (2D)
        :param y_train: labels (1D)
        :return: None
        """
        self.inner_cython_mnb.fit(x_train, y_train)
    def predict(self, x_test):
        """
        :param x_test: testing data (2D)
        :return: predictions in a 1D vector/list
        """
        return np.array(self.inner_cython_mnb.predict(x_test))
    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: Just what percent of predictions on x_test are actually correct
        """
        return self.inner_cython_mnb.evaluate(x_test, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()

class GaussianNaiveBayes() :
    """
    Gaussian Naive Bayes is the lesser known version of Naive Bayes. It is used for the standard numeric data you may
    feed into a model like logistic regression. What it does is create a normal (Gaussian) probability distribution for each
    feature so it can measure the chance that a data point belongs to certain class given that it has this or that feature.
    It will do this for every data point, for every class, and for every feature (not as efficient as Softmax Regression.)
    Just another cool algorithm.

    ----
    Methods :

    __init__() :
        ->> no params here
    fit(x_train, y_train) :
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)
    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> will return all predictions in a 1-D vector/python list.
    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> will the percent of predictions on x_test it got correct
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """
    def __init__(self):
        from .cython_naive_bayes import cy_GaussianNaiveBayes
        self.inner_cython_gnb = cy_GaussianNaiveBayes()
    def fit(self, x_train, y_train):
        """
        :param x_train: training data (2D)
        :param y_train: training labels (1D)
        :return:
        """
        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")
        self.inner_cython_gnb.fit(x_train, y_train)
    def predict(self, x_test):
        """
        :param x_test: prediction data (2D)
        :return: predictions in a 1D vector/list
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return np.array(self.inner_cython_gnb.predict(x_test))
    def evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")
        return self.inner_cython_gnb.evaluate(x_test, y_test)
    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()


class BernoulliNaiveBayes() :
    """
    Bernoulli Naive Bayes is a special type of naive bayes, that's not nearly as common as the others. It is meant
    for binary classification (is it spam (1) or nonspam (0)?) only. Unlike multinomial naive bayes, what bernoulli naive
    bayes does is look for not just the presence of a feature (i.e. a word) to predict whether something is one class
    or another - it also looks for the absence of a feature to make that decision. This means that Bernoulli
    Naive Bayes has to iterate through each word in its vocabulary, so it can be slightly slower. It also gets it's
    vocabulary from the training data, so (as with most of the naive bayes models) having good data that's representative
    and balanced will help.

    ----
    Methods :

    __init__() :
        ->> no params here
    fit(x_train, y_train) :
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)
    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> will return all predictions in a 1-D vector/python list.
    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> will the percent of predictions on x_test it got correct
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """
    def __init__(self):
        from .cython_naive_bayes import cy_BernoulliNaiveBayes
        self.inner_cython_bnb = cy_BernoulliNaiveBayes()
    def fit(self, x_train, y_train):
        """
        :param x_train: training data (2D)
        :param y_train: training labels (1D)
        :return:
        """
        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")
        self.inner_cython_bnb.fit(x_train, y_train)
    def predict(self, x_test):
        """
        :param x_test: prediction data (2D)
        :return: predictions in a 1D vector/list
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return np.array(self.inner_cython_bnb.predict(x_test))
    def evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")
        return self.inner_cython_bnb.evaluate(x_test, y_test)
    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()





"""
@author : Anish Lakkapragada
@date : 1 - 7 - 2021

This utils module is for all the "other stuff" used in ML. Here we provide a function to plot a confusion matrix,
a one-hot encoder, and a reverse one-hot encoder too.
"""
import numpy as np
import pandas as pd


def confusion_matrix(y_pred, y_test, plot=True):
    """

    Confusion matrices are an often used tool for seeing how well your model did and where it can improve.

    A confusion matrix is just a matrix with the number of a certain class classified as another. So if your classifier
    predicted [0, 1] and the correct answers were [1, 0] - then you would get 1 zero predicted as a 1, 1 one predicted
    as a 0, and no 0s predicted as 0s and 1s predicted as 1s. By default this method will plot the confusion matrix, but
    if you don't want it just set it to False.

    Some warnings - make sure that y_pred and y_test are both 1D and start at 0. Also make sure that all predictions in
    y_pred exist in y_test. This will probably not be a big deal with traditional datasets for machine learning.

    To really understand this function try it out - it'll become clear with the visualization.

    :param y_pred: predictions (1D)
    :param y_test: labels (1D)
    :param plot: whether or not to plot, default True
    :return: the matrix, and show a visualization of the confusion matrix (if plot = True)
    """

    y_pred, y_test = (
        np.array(y_pred, np.int).tolist(),
        np.array(y_test, np.int).tolist(),
    )

    if len(y_pred) != len(y_test):
        raise ValueError("y_pred and y_test must have the same size.")

    possible_classes = set(y_test)
    if 0 not in possible_classes:
        raise ValueError("Labels must be start at 0.")

    confusion_matrix_dict = {
        lbl: {cls: 0 for cls in possible_classes} for lbl in possible_classes
    }

    for pred, label in zip(y_pred, y_test):
        if pred not in y_test:
            raise ValueError(
                "All predictions must belong to classes that are in y_test."
            )
        confusion_matrix_dict[label][pred] += 1

    conf_matrix = np.identity(
        len(confusion_matrix_dict)
    )  # start the matrix off with an identity matrix
    for label, pred_dict in confusion_matrix_dict.items():
        """label is the row, keys of pred_dict are the cols, values of the dict are the values"""
        for prediction, amount in pred_dict.items():
            conf_matrix[label][prediction] = amount

    if plot:
        import seaborn as sns
        import matplotlib.pyplot as plt

        plt.clf()
        plt.cla()
        labels = np.identity(len(possible_classes)).tolist()
        for row in range(len(labels)):
            for col in range(len(labels[row])):
                if row == col:
                    """correct diagonal"""
                    labels[row][col] = str(conf_matrix[row][col])
                else:
                    """wrong diagonal"""
                    """correct diagonal"""
                    labels[row][col] = str(conf_matrix[row][col])
        labels = np.asarray(labels).reshape(
            len(possible_classes), len(possible_classes)
        )
        sns.heatmap(conf_matrix, annot=labels, fmt="", cmap="Blues")
        plt.title("Confusion Matrix")
        plt.xlabel("Predicted")
        plt.ylabel("Labels")
    return conf_matrix


def _change(zeros_and_index):
    zeros, index = zeros_and_index
    # print("This is zeros : ", zeros)
    # print("This is index : ", index)
    zeros[index] = 1
    return zeros


def one_hot(indices, depth):
    """
    So you've got a feature in a data where a certain value represents a certain category. For example it could be
    that 1 represents sunny, 2 represents rainy, and 0 represents cloudy. Well what's the problem? If you feed this
    to your model - it's going to think that 2 and 1 are similar, because they are just 1 apart - despite the fact
    that they are really just categories. To fix that you can feed in your features - say it's a list like [2, 2, 1,
    0, 1, 0] and it will be one hot encoded to whatever depth you please.

    Here's an example (it'll make it very clear) :
    >>> features_weather = [1, 2, 2, 2, 1, 0]
    >>> one_hot_features = one_hot(features_weather, depth  = 3) #depth is three here because you have 3 categories - rainy, sunny, cloudy
    >>> one_hot_features
    [[0, 1, 0], #1 is at first index
     [0, 0, 1], #2 is at second index
     [0, 0, 1], #2 is at second index
     [0, 0, 1], #2 is at second index
     [0, 1, 0], #1 is at first index
     [1, 0, 0]] #0 is at 0 index

    That looks like our data features, one hot encoded at whatever index. Make sure to set the depth param correctly.

    For these such things, play around - it will help.

    :param indices: features_weather or something similar as shown above
    :param depth: How many categories you have
    :return: one-hotted features
    """

    indices = np.array(indices, dtype=np.int).flatten()
    zeros = np.zeros((len(indices), depth))
    df = pd.DataFrame({"col1": zeros.tolist(), "col2": indices.tolist()})
    return np.apply_along_axis(_change, 1, df)


def revert_one_hot_nn(one_hot_data):
    """
    Say from the one_hot() data you've gotten something like this :

    [[0, 0, 1],
    [1, 0, 0],
    [0, 1, 0],
    [0, 0, 1],
    [1, 0, 0]]

    and you want to change it back to its original form. Use this function - it will turn that one-hotted data above to
    [2, 0, 1, 2, 0] - which is just the index of the one in each data point of the list. Hence it is reverting the
    one_hot transformation.

    :param one_hot_data: data in one_hot form. Must be a numpy array (2D)!
    :return: index of the one for each of the data points in a 1D numpy array
    """

    if np.isnan(one_hot_data).any():
        raise ValueError("No NaN values are allowed.")

    if not float(np.sum(one_hot_data)) == float(one_hot_data.shape[0]):
        raise ValueError("Data isn't one-hotted encoded properly.")

    return np.apply_along_axis(np.argmax, 1, one_hot_data)


def revert_softmax_nn(softmax_data):
    """
    Say from the softmax function (in neural networks) you've gotten something like this :

    [[0.2, 0.3, 0.5],
    [0.8, 0.1, 0.1],
    [0.15, 0.8, 0.05],
    [0.3, 0.3, 0.4],
    [0.7, 0.15, 0.15]]

    and you want to change it back to its pre-softmax form. Use this function - it will turn that softmax-ed data above to
    [2, 0, 1, 2, 0] - which is just the index of the one in each data point of the list. Hence it is reverting the
    softmax transformation.

    :param softmax_data: data in one_hot form. Must be a numpy array (2D)!
    :return: index of the one for each of the data points in a 1D numpy array
    """

    if np.isnan(softmax_data).any():
        raise ValueError("No NaN values are allowed.")

    return np.apply_along_axis(np.argmax, 1, softmax_data)



"""
@author : Anish Lakkapragada
@date : 1 - 23 - 2021

This file contains all of the layer classes you'll need to build your models. You won't need the forward() and
backward() methods to deal with, just really the init. Examples are embedded in the docs because this can get a little complex,
so bear with me here.
"""
import numpy as np


class Layer:
    def __init__(self):
        self.parameters = {}
        self.gradients = {}
        self.activation = None
        self.nesterov = False
        self.indices = ...
        pass

    def forward(self, inputs):
        """This method saves the inputs, and returns its outputs"""
        pass

    def backward(self, grad):
        """This method takes in a gradient (e.x ∂l/∂Z2) and returns its grad (e.x. dL/dA1)"""

        pass


class Flatten(Layer):
    """
    This would be better explained as Image Data Flattener. Let's say your dealing with MNIST (check out the examples
    on github) - you will have data that is 60000 by 28 by 28. Neural networks can't work with data like that - it has
    to be a matrix. What you could do is take that 60000 * 28 * 28 data and apply this layer to make the data
    60000 * 784. 784 because 28 * 28 = 784 and we just "squished" this matrix to one layer. If you have data that has colors,
    it maybe like 60000 * 28 * 28 * 3 so applying this layer would turn it into a 60000 * 2352 matrix. 2352 as 28 * 28 * 3
    is 2352 here.

    An example for how this would work with something like MNIST (the grayscale 60k * 28 * 28 dataset) is shown below.

    That would look like :
    >>> from sealion import neural_networks as nn
    >>> from sealion.neural_networks.models import NeuralNetwork
    >>> model = NeuralNetwork()
    >>> model.add(nn.layers.Flatten()) #always, always add that on the first layer
    >>> model.add(nn.layers.Dense(784, whatever_output_size, more_args)) #now put that data to 784 as the 28 * 28 is flattened to just 784
    >>> Do more cool stuff...
    """

    def forward(self, inputs):
        self.inputs = inputs
        return inputs.reshape(inputs.shape[0], np.product(inputs.shape[1:]))

    def backward(self, grad):
        return self.inputs


class Dropout(Layer):
    """
    Dropout is one of the most well-known regularization techniques there is. Imagine you were working on a coding
    project with about 200 people. If we just relied on one person to know how to compile, another person to know how to debug,
    then what happens when those special people leave for a day, or worse leave forever?

    Now I know that seems to have no connection to dropout, but here's how it does. In dropout this sort of scenario is prevented.
    "Dropping out", or setting to 0, some of the outputs of some of the neurons means that the model will have to learn
    that it can't just depend on one neuron to learn the most important features. This means has each neuron learn some features,
    some other features, etc. and can't just depend on one node. The model will become more robust and generalize
    better with Dropout as every neuron know has a better set of weights. Normally due to dropout, it will be applied
    in training, but then "reverse-applied" in testing. Dropout will make the training accuracy go down a bit, but remember
    in the end it's testing on real-world data that matters.

    There's a parameter dropout_rate on what percent (from 0 - 1 here) you want each neuron in the layer you are at to become 0.
    This is essentially the chance of dropping out any given neuron, or usually what percent of neurons will be dropped out. Typical
    values range from 0.1 - 0.5. Example below.

    Let's say you've gotten your models up so far :
    >>> model.add(nn.layers.Flatten())
    >>> model.add(nn.layers.Dense(128, 64, activation = nn.layers.ReLU()))

    And now you want to add dropout. Well just add that layer like such :
    >>> dropout_probability = 0.2
    >>> model.add(nn.layers.Dropout(dropout_probability))

    This will just mean that about 20% of the neurons coming from this first input layer will be dropped out. A higher dropout rate
    may not always lead to better generalization, but usually will decrease training accuracy.

    In dropout remember 2 things, not just one matter. The probability, and the layers its applied at. Experimentation is key.
    """

    def __init__(self, dropout_rate):
        super().__init__()
        self.dr = dropout_rate

    def forward(self, inputs):
        self.inputs = inputs
        self.dropped_inputs = self.inputs * np.random.binomial(
            1, (1 - self.dr), inputs.shape
        )
        return self.dropped_inputs

    def backward(self, grad):
        grad[np.where(self.dropped_inputs[self.indices] == 0)] = 0 # set droppped out outputs to 0 
        grad *= 1/(1-dropout_rate) # everything else scale by * 1/(1-p)
        return grad


class Dense(Layer):
    """

    This is the class that you will depend on the most. This class is for creating the fully-connected layers that
    make up Deep Neural Networks - where each neuron in a previous layer is connected to each layer in the next. Feel free
    to watch a couple of youtube tutorials from 3Blue1Brown (if you like calculus :) or others to get a better understanding
    of these parameters, or just look at the examples on my github.

    The main method of course is the init. You will have to define the input_size, the output_size, and the activation and
    weight initialization are optional parameters.

    In a neural network the number of nodes starting in a layer are known as the input_size here (fan_in in papers), and
    the number of nodes in the output of a layer is the output_size here (fan_out in papers.) The activation parameter
    is for the activation/non-linearity function you want your layers to go through. If you don't understand what I just meant,
    sorry - but another way to think about it is that its a way to change the outputs of your network a bit for it to
    fit you dataset a little better (looking at a graph will help.) The default is no activation (some people call that Linear),
    so you'll need to add that yourself. I'll get to weight init in a bit. Examples below.

    To add a layer, here's how it's done :
    >>> from sealion import neural_networks as nn
    >>> model = nn.models.NeuralNetwork()
    >>> model.add(nn.layers.Dense(128, 64)) #input_size : 128, output_size : 64

    This sets up a neural network with 128 incoming nodes (that means we have 128 numeric features), and then 64 output nodes.

    Let's say we wanted an activation function, like a relu or sigmoid (there are a bunch to choose from this API.)
    You could add that layer here like such :

    >>> model = nn.models.NeuralNetwork() #clear all existing layers
    >>> model.add(nn.layers.Dense(128, 64, activation = nn.layers.Sigmoid())) #all outputs go through the sigmoid function

    Onto weight initalization! The jargon starts .... now. What weight init does is make the weights come from
    a special distribution (typically Gaussian) with a given standard deviation based on the input and output size.
    The reason this is done is because you don't want the weights to be initalized too big or else the gradients in
    backpropagation may cause the model to go way off and become NaNs or infinity (this is known as exploding gradients.)
    For neural networks that are small that solve datasets like XOR or even breast cancer this isn't a problem, but for
    deep neural networks for problems like MNIST this is a huge concern. The weight_init you will want to use is also
    dependent on what activation you are using. Most activation functions will do well with Xavier Glorot, so that is
    set as the default. You can choose to also use He for ReLU, LeakyReLU, ELU, or other variants of the ReLU activation.
    For SELU, you may choose to use LeCun weight initalization.

    The possible choices are :
    >>> "xavier" #for no activation, logistic/sigmoid, softmax, and tanh
    >>> "he" #for relu + similar variants
    >>> "lecun" #for selu
    >>> "none" # if you want to do this

    To set this you can just do :
    >>> model = nn.models.NeuralNetwork() #clear all existing layers
    >>> model.add(nn.layers.Dense(128, 64, activation = nn.layers.ReLU(), weight_init = "he"))

    Sorry for so much documentation, but this really is the class you will call the most.
    """

    def __init__(
        self, input_size: int, output_size: int, activation=None, weight_init="xavier"
    ) -> None:
        super().__init__()

        weight_init_addition = 1  # no weight init
        if weight_init.lower() == "xavier":
            weight_init_addition = np.sqrt(2 / (output_size + input_size))
        elif weight_init.lower() == "he":
            weight_init_addition = np.sqrt(2 / input_size)
        elif weight_init.lower() == "lecun":
            weight_init_addition = np.sqrt(1 / input_size)
        elif not weight_init.lower() == "none":
            print(f"No known weight init : {weight_init}")

        self.parameters["weights"] = (
            np.random.randn(input_size, output_size) * weight_init_addition
        )
        self.parameters["bias"] = np.random.randn(output_size)
        self.gradients["weights"] = np.random.randn(*self.parameters["weights"].shape)
        self.gradients["bias"] = np.random.randn(*self.parameters["bias"].shape)
        self.activation = activation

    def forward(self, inputs):
        self.inputs = inputs
        return np.dot(inputs, self.parameters["weights"]) + self.parameters["bias"]

    def backward(self, grad):
        # here we can update our weights
        """if gradient ∂L/∂ZN, store grads ∂l/∂w1 and return dL/dA1"""

        self.gradients["weights"] = self.inputs[self.indices].T.dot(grad)
        self.gradients["bias"] = np.sum(grad, axis=0)
        return np.dot(grad, (self.parameters["weights"]).T)  # dLdZ2 * dZ2/dA1


class Activation(Layer):
    """
    To add Activation Functions, check the Dense class for a tutorial, or the examples.
    """

    def __init__(self, activ_func, activ_func_prime):
        super().__init__()
        self.activ_func = activ_func
        self.activ_func_prime = activ_func_prime

    def forward(self, inputs):
        self.inputs = inputs
        return self.activ_func(inputs)

    def backward(self, grad):
        return self.activ_func_prime(self.inputs[self.indices]) * grad


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def sigmoid_prime(z):
    return sigmoid(z) * (1 - sigmoid(z))


def tanh(x):
    return np.tanh(x)


def tanh_prime(z):
    return 1 - np.power(tanh(z), 2)


def relu(x):
    return np.maximum(x, 0)


def relu_prime(z):
    return (z > 0).astype(int)


def leaky_relu(x, leak=0.01):
    return np.maximum(leak * x, x)


def leaky_relu_prime(z, leak=0.01):
    grad = np.ones_like(z)
    grad[z < 0] *= leak
    return grad


def elu(z, alpha=1):
    return np.where(z > 0, z, alpha * (np.exp(z) - 1))


def elu_prime(z, alpha=1):
    grad = np.where(z > 0, np.ones_like(z), alpha * (np.exp(z)))
    return grad


def swish(x):
    return x * sigmoid(x)


def swish_prime(z):
    return swish(z) + sigmoid(z) * (1 - swish(z))


def selu(x):
    alpha = 1.6732
    lamb = 1.0507
    return lamb * np.maximum((alpha * np.exp(x) - alpha), x)


def selu_prime(x):
    alpha = 1.6732
    lamb = 1.0507
    grad = np.where(x <= 0, np.ones_like(x), lamb * alpha * np.exp(x))
    return grad


class Tanh(Activation):
    """Uses the tanh activation, which squishes values from -1 to 1."""

    def __init__(self):
        super().__init__(tanh, tanh_prime)


class Sigmoid(Activation):
    """Uses the sigmoid activation, which squishes values from 0 to 1."""

    def __init__(self):
        super().__init__(sigmoid, sigmoid_prime)


class Swish(Activation):
    """
    Uses the swish activation, which is sort of like ReLU and sigmoid combined. It's really just f(x) = x * sigmoid(x).
    Not as used as other activation functions, but give it a try!
    """

    def __init__(self):
        super().__init__(swish, swish_prime)


class ReLU(Activation):
    """
    The most well known activation function and the pseudo-default almost. All it does is turn negative values to 0 and
    keep the rest.
    """

    def __init__(self):
        super().__init__(relu, relu_prime)


class LeakyReLU(Activation):
    """
    Variant of the ReLU activation, just allows negatives values to be something more like 0.01 instead of 0, which means
    the neuron is "dead" as 0 * anything is 0.

    The leak is the slope of how low negative values you are willing to tolerate. Usually set from 0.001 - 0.2, but the
    default of 0.01 works usually quite well.
    """

    def __init__(self, leak=0.01):
        super().__init__(leaky_relu, leaky_relu_prime)
        self.leak = leak

    def forward(self, inputs):
        self.inputs = inputs
        return self.activ_func(inputs, leak=self.leak)

    def backward(self, grad):
        return self.activ_func_prime(self.inputs[self.indices], leak=self.leak) * grad


class ELU(Activation):
    """
    Solves the similar dying activation problem. The default of 1 for alpha works quite well in practice, so you won't need
    to change it much.
    """

    def __init__(self, alpha=1):
        super().__init__(elu, elu_prime)
        self.alpha = alpha

    def forward(self, inputs):
        self.inputs = inputs
        return self.activ_func(inputs, alpha=self.alpha)

    def backward(self, grad):
        return self.activ_func_prime(self.inputs[self.indices], alpha=self.alpha) * grad


class SELU(Activation):
    """
    Special type of activation function, that will "self-normalize" (have a mean of 0, and a standard deviation of 1) its outputs.
    This self-normalization typically leads to faster convergence.

    If you are using this activation function make sure weight_init is set = "lecun" in whichever layer applied.
    It also need its inputs (in the beginning and all throughout) to be standardized (mu = 0, sigma = 1) for it to work,
    so make sure to get that taken care of. You can do that by standardizing your inputs and then always using
    SELU activation function (remember the "lecun" part!).
    """

    def __init__(self):
        super().__init__(selu, selu_prime)

class PReLU(Layer) :
    """
    The PReLU activation function is essentially the same thing as the LeakyReLU activation, except that the "leak"
    parameter is learnt during training. To learn this parameter, gradient descent is used - so you have a learning rate and momentum parameter.
    Both are on a scale of 0 - 1. We initialize this "leak" parameter to 0.25 at the first iteration.
    """

    def __init__(self, lr = 0.0001, momentum = 0.9) :
        super().__init__()
        self.a = 0.25
        self.delta_a = 0 
        self.momentum = momentum
        self.lr = lr
    def forward(self, inputs) :
        self.inputs = inputs
        self.outputs = np.maximum(inputs, 0) + self.a * np.minimum(inputs, 0)
        return self.outputs
    def backward(self, grad) :
        # gradient is the ∂J/∂Output of this PReLU function
        
        # first find ∂J/∂a and update a
        dOutputda = np.zeros(self.outputs.shape)
        dOutputda[np.where(self.outputs < 0)] = self.inputs[np.where(self.outputs < 0)]

        # now return back dJdInput
        dOutputdInput = np.ones(self.inputs.shape)
        dOutputdInput[np.where(self.outputs <= 0)] = self.a
        
        self.delta_a = self.momentum * self.delta_a + self.lr * (grad * dOutputda).sum()
        self.a += self.delta_a  
 
        return grad * dOutputdInput
    

class Softmax(Layer):
    """
    Softmax activation function, used for multi-class (2+) classification problems.
    Make sure to use crossentropy with softmax, and it is only meant for the last layer!
    """

    def forward(self, inputs):
        self.inputs = inputs




"""
@author : Anish Lakkapragada
@date : 1 - 23 - 2021

The optimizer APIs really go from the simple basic Gradient Descent optimization algorithms to newer, and more nuanced
algorithms like Adam and RMSprop. The optimizers will build up in sequential order, so please read the documentation in
that order too if you are unfamiliar with these algorithms.

You may want to check out this blog post to understand these algorithms in more detail :
https://ruder.io/optimizing-gradient-descent/

"""

import numpy as np


class Optimizer:
    def __init__(self):
        self.nesterov = False

    def setup(self, nn):
        pass

    def update(self, nn):
        pass


class GD(Optimizer):
    """

    The simplest optimizer -  you will quickly outgrow it. All you need to understand here is that the learning rate
    is just how fast you want the model to learn (default 0.001) set typically from 1e-6 to 0.1. A higher learning rate may mean
    the model will struggle to learn, but a slower learning rate may mean the model will but will also have to take more
    time. It is probably the most important hyperparameter in all of today's machine learning.

    The clip_threshold is simply a value that states if the gradients is higher than this value, set it to this.
    This is to prevent too big gradients, which makes training harder. The default for all these optimizers is infinity,
    which just means no clipping - but feel free to change that. You'll have to experiment quite a bit to find a good value.
    This method is known as gradient clipping.

    In the model.finalize() method :
    >>> model.finalize(loss = ..., optimizer = nn.optimizers.GD(lr = 0.5, clip_threshold = 5)) # here the learning
    # learning rate is 0.5 and the threshold for clipping is 5.
    """

    def __init__(self, lr=0.001, clip_threshold=np.inf):
        super().__init__()
        self.lr = lr
        self.clip_threshold = clip_threshold

    def update(self, nn):
        for layer in nn:
            for param_name, _ in layer.parameters.items():
                gradients = layer.gradients[param_name]
                gradients[
                    np.where(np.abs(gradients) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] -= self.lr * gradients


class Momentum(Optimizer):
    """
    If you are unfamiliar with gradient descent, please read the docs on that in GD class for this to hopefully make more sense.

    Momentum optimization is the exact same thing except with a little changes. All it does is accumulate the past gradients,
    and go in that direction. This means that as it makes updates it gains momentum and the gradient updates become bigger and bigger
    (hopefully in the right direction.) Of course this will be uncontrolled on its own, so a momentum parameter (default 0.9)
    exists so the previous gradients sum don't become too big. There is also a nesterov parameter (default false, but
    set that true!) which sees how the loss landscape will be in the future, and makes its decisions based off of that.

    An example :
    >>> momentum = nn.optimizers.Momentum(lr = 0.02, momentum = 0.3, nesterov = True) #learning rate is 0.2, momentum at 0.3, and we have nesterov!
    >>> model.finalize(loss = ..., optimizer = momentum)

    There's also a clip_threshold argument which you implements gradient clipping, an explanation you can find in the GD()
    class's documentation.

    Usually though this works really good with SGD...
    """

    def __init__(self, lr=0.001, momentum=0.9, nesterov=False, clip_threshold=np.inf):
        super().__init__()
        self.lr = lr
        self.momentum = momentum
        self.nesterov = nesterov
        self.clip_threshold = clip_threshold

    def setup(self, nn):
        self.momentum_params = {}
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )

        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):
        """nn is self.layers"""
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] = (
                    self.momentum * self.momentum_params[param_name + str(layer_num)]
                    - self.lr * layer.gradients[param_name]
                )
                final_grad = self.momentum_params[param_name + str(layer_num)]
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] += final_grad


class SGD(Optimizer):
    """
    SGD stands for Stochastic gradient descent, which means that it calculates its gradients on random (stochastic) picked samples
    and their predictions. The reason it does this is because calculating the gradients on the whole dataset can take a
    really long time. However ML is a tradeoff and the one here is that calculating gradients on just a few samples
    means that if those samples are all outliers it can respond poorly, so SGD will train faster but not get as high an
    accuracy as Gradient Descent on its own.

    Fortunately though, there are work arounds. Implementing momentum and nesterov with SGD means you get faster training
    and also the convergence is great as now the model can go in the right direction and generalize instead of overreact
    to hyperspecific training outliers. By default nesterov is set to False and there is no momentum (set to 0.0), so please
    change that as you please.

    To use this optimizer, just do :
    >>> model.finalize(loss = ..., optimizer = nn.optimizers.SGD(lr = 0.2, momentum = 0.5, nesterov = True, clip_threshold = 50))

    Here we implemented SGD optimization with a learning rate of 0.2, a momentum of 0.5 with nesterov's accelerated gradient, and
    also gradient clipping at 50.

    """

    def __init__(self, lr=0.001, momentum=0.0, nesterov=False, clip_threshold=np.inf):
        super().__init__()
        self.lr = lr
        self.momentum = momentum
        self.nesterov = nesterov
        self.clip_threshold = clip_threshold

    def setup(self, nn):
        """nn is self.layers"""
        self.momentum_params = {}
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )

        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):
        """nn is self.layers"""
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] = (
                    self.momentum * self.momentum_params[param_name + str(layer_num)]
                    - self.lr * layer.gradients[param_name]
                )
                final_grad = self.momentum_params[param_name + str(layer_num)]
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] += final_grad


class AdaGrad(Optimizer):
    """
    Slightly more advanced optimizer, an understanding of momentum will be invaluable here. AdaGrad and a whole plethora
    of optimizers use adaptive gradients, or an adaptive learning rate. This just means that it will assess the landscape
    of the cost function, and if it is steep it will slow it down, and if it is flatter it will accelerate. This is a huge
    deal for avoiding gradient descent from just going into a steep slope that leads to a local minima and being stuck,
    or gradient descent being stuck on a saddle point.

    The only new parameter is e, or this incredibly small value that is meant to prevent division by zero. It's set to
    1e-10 by default, and you probably won't ever need to think about it.

    As an example :
    >>> model.finalize(loss = ..., optimizer = nn.optimizers.AdaGrad(lr = 0.5, nesterov = True, clip_threshold = 5))

    AdaGrad is not used in practice much as often times it stops before reaching the global minima due to the gradients
    being too small to make a difference, but we have it anyways for your enjoyment. Better optimizers await!

    """

    def __init__(self, lr=0.001, nesterov=False, clip_threshold=np.inf, e=1e-10):
        super().__init__()
        self.lr = lr
        self.nesterov = nesterov
        self.clip_threshold = clip_threshold
        self.e = e

    def setup(self, nn):
        self.momentum_params = {}  # the s in page 354 of handsonmlv2
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] += np.power(
                    layer.gradients[param_name], 2
                )
                final_grad = (
                    self.lr
                    * layer.gradients[param_name]
                    / (
                        np.sqrt(self.momentum_params[param_name + str(layer_num)])
                        + self.e
                    )
                )
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] -= final_grad


class RMSProp(Optimizer):
    """
    RMSprop is a widely known and used algorithm for deep neural network. All it does is solve the problem AdaGrad has
    of stopping too early by not scaling down the gradients so much. It does through a beta parameter, which is set to 0.9
    (does quite well in practice.) A higher beta means that past gradients are more important, and a lower one means current
    gradients are to be valued more.

    An example :
    >>> model.finalize(loss = ..., optimizer = nn.optimizers.RMSprop(nesterov = True, beta = 0.9))

    Of course there is the nesterov, clipping threshold, and e parameter all for you to tune.
    """

    def __init__(
        self, lr=0.001, beta=0.9, nesterov=False, clip_threshold=np.inf, e=1e-10
    ):
        super().__init__()
        self.lr = lr
        self.beta = beta
        self.nesterov = nesterov
        self.clip_threshold = clip_threshold
        self.e = e

    def setup(self, nn):
        self.momentum_params = {}  # the s in page 354 of handsonmlv2
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_params[
                    param_name + str(layer_num)
                ] = self.beta * self.momentum_params[param_name + str(layer_num)] + (
                    1 - self.beta
                ) * np.power(
                    layer.gradients[param_name], 2
                )
                final_grad = (
                    self.lr
                    * layer.gradients[param_name]
                    / (
                        np.sqrt(self.momentum_params[param_name + str(layer_num)])
                        + self.e
                    )
                )
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] -= final_grad


class Adam(Optimizer):
    """
    Most popularly used optimizer, typically considered the default just like ReLU for activation functions.
    Combines the ideas of RMSprop and momentum together, meaning that it will adapt to different landscapes but move
    in a faster direction. The beta1 parameter (default 0.9) controls the momentum optimization,
    and the beta2 parameter (default 0.999) controls the adaptive learning rate parameter here. Once again higher betas
    mean the past gradients are more important.

    Often times you won't know what works best - so hyperparameter tune.

    As an example :
    >>> model.finalize(loss = ..., optimizer = nn.optimizers.Adam(lr = 0.1, beta1 = 0.5, beta2 = 0.5))

    Adaptive Gradients may not always work as good as SGD or Nesterov + Momentum optimization. For MNIST, I have tried
    both and there's barely a difference. If you are using Adam optimization and it isn't working maybe try using
    nesterov with momentum instead.
    """

    def __init__(
        self,
        lr=0.001,
        beta1=0.9,
        beta2=0.999,
        nesterov=False,
        clip_threshold=np.inf,
        e=1e-10,
    ):
        super().__init__()
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.nesterov = nesterov
        self.clip_threshold = clip_threshold
        self.e = e
        self.t = 1

    def setup(self, nn):
        self.momentum_M, self.momentum_S = {}, {}
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_M[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
                self.momentum_S[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):

        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_M[param_name + str(layer_num)] = (
                    self.beta1 * self.momentum_M[param_name + str(layer_num)]
                    - (1 - self.beta1) * layer.gradients[param_name]
                )
                self.momentum_S[
                    param_name + str(layer_num)
                ] = self.beta2 * self.momentum_S[param_name + str(layer_num)] + (
                    1 - self.beta2
                ) * np.power(
                    layer.gradients[param_name], 2
                )
                m_hat = self.momentum_M[param_name + str(layer_num)] / (
                    1 - np.power(self.beta1, self.t)
                )
                s_hat = self.momentum_S[param_name + str(layer_num)] / (
                    1 - np.power(self.beta2, self.t)
                )
                final_grad = self.lr * m_hat / (self.e + np.sqrt(s_hat))
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] += final_grad


class Nadam(Optimizer):
    """
    Nadam optimization is the same thing as Adam, except there's nesterov updating. Basically this class is the same as
    the Adam class, except there is no nesterov parameter (default true.)

    As an example :
    >>> model.finalize(loss = ..., optimizer = nn.optimizers.Nadam(lr = 0.1, beta1 = 0.5, beta2 = 0.5))
    """

    def __init__(
        self, lr=0.001, beta1=0.9, beta2=0.999, clip_threshold=np.inf, e=1e-10
    ):
        super().__init__()
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.nesterov = True
        self.clip_threshold = clip_threshold
        self.e = e
        self.t = 1

    def setup(self, nn):
        self.momentum_M, self.momentum_S = {}, {}
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_M[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
                self.momentum_S[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_M[param_name + str(layer_num)] = (
                    self.beta1 * self.momentum_M[param_name + str(layer_num)]
                    - (1 - self.beta1) * layer.gradients[param_name]
                )
                self.momentum_S[
                    param_name + str(layer_num)
                ] = self.beta2 * self.momentum_S[param_name + str(layer_num)] + (
                    1 - self.beta2
                ) * np.power(
                    layer.gradients[param_name], 2
                )
                m_hat = self.momentum_M[param_name + str(layer_num)] / (
                    1 - np.power(self.beta1, self.t)
                )
                s_hat = self.momentum_S[param_name + str(layer_num)] / (
                    1 - np.power(self.beta2, self.t)
                )
                final_grad = self.lr * m_hat / (np.sqrt(s_hat) + self.e)
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.gradients[param_name] += final_grad


class AdaBelief(Optimizer):
    """
    AdaBelief is a recently popular optimizer that I thought to implement
    after checking in with the original paper. The way it works is by establishing the
    general "belief" (prediction) in where the gradient will go. If the belief and the actual 
    given gradient are very similar, there will be a large change in the weights
    (larger gradients), whereas if they are very different, there will be a small 
    change in the weights. This has worked well in practice, and its results
    are comparable to Adam. 

    For those interested in the original paper, go here : https://arxiv.org/pdf/2010.07468.pdf
    """
    
    def __init__(
        self,
        lr=0.001,
        beta1=0.9,
        beta2=0.999,
        nesterov=False,
        clip_threshold=np.inf,
        e=1e-10,
    ):
        super().__init__()
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.nesterov = nesterov
        self.clip_threshold = clip_threshold
        self.e = e
        self.t = 1

    def setup(self, nn):
        self.momentum_M, self.momentum_S = {}, {}
        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_M[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
                self.momentum_S[param_name + str(layer_num)] = np.zeros(
                    layer.gradients[param_name].shape
                )
        if self.nesterov:
            for layer_num in range(len(nn)):
                layer = nn[layer_num]
                layer.nesterov = True
                nn[layer_num] = layer

    def update(self, nn):

        for layer_num in range(len(nn)):
            layer = nn[layer_num]
            for param_name, _ in layer.parameters.items():
                self.momentum_M[param_name + str(layer_num)] = (
                    self.beta1 * self.momentum_M[param_name + str(layer_num)]
                    - (1 - self.beta1) * layer.gradients[param_name]
                )
                self.momentum_S[param_name + str(layer_num)] = self.beta2 * self.momentum_S[param_name + str(layer_num)] + (1 - self.beta2) * np.power(
                    layer.gradients[param_name] - self.momentum_M[param_name + str(layer_num)], 2
                ) + self.e

                m_hat = self.momentum_M[param_name + str(layer_num)] / (
                    1 - np.power(self.beta1, self.t)
                )
                s_hat = self.momentum_S[param_name + str(layer_num)] / (
                    1 - np.power(self.beta2, self.t)
                )
                final_grad = self.lr * m_hat / (self.e + np.sqrt(s_hat))
                final_grad[
                    np.where(np.abs(final_grad) > self.clip_threshold)
                ] = self.clip_threshold
                layer.parameters[param_name] += final_grad



"""
@author : Anish Lakkapragada
@date : 3 - 4 - 2021

Gaussian Mixtures are one of the best models today in the field of anomaly detection and unsupervised clustering.
But it's not just because of that that this is one of the best modules SeaLion has to offer.
"""

import numpy as np

class GaussianMixture() :
    """
    Gaussian Mixture Models are really just a fancier extension of KMeans. Make sure you know KMeans before you read this.
    If you are unfamiliar with KMeans, feel free to look at the examples on GitHub for unsupervised clustering or
    look at the unsupervised clustering documentation (which contains the KMeans docs.) You may also want to know
    what a gaussian (normal) distribution is.

    In KMeans, you make the assumption that your data is in spherical clusters, all of which are the same shape. What
    if your data is instead in such circles, but also maybe ovals that are thin, tall, etc. Basically what if your
    clusters are spherical-esque but of different shapes and sizes.

    This difference can be measured by the standard deviation, or variance, of each of the clusters. You can think of each
    cluster as a gaussian distribution, each with a different mean (similiar to the centroids in KMeans) and standard
    deviation (which effects how skinny/wide it is.) This model is called a mixture, because it essentially is just a
    mixture of gaussian functions. Some of the clusters maybe bigger than others (aside from shape), and this is
    also taken into account with a mixture weight - a coefficient that is multiplied to each gaussian distribution.

    With this gaussian distribution, you can then take any points and assign them to the cluster they have the
    highest change of being in. Because this is probabilities, you can say that a given data point has a 70% chance
    of being in this class, 20% this class, 5% this class, and 5% this other class - which is something you
    cannot do with KMeans. The parameters of the gaussian distribution are learnt using an algorithm known
    as Expectation Maximization, which you may want to learn about (super cool)!

    You can also do anomaly detection with Gaussian Mixture Models! You do this by looking at the probability each data
    point belonged to the cluster it had the highest chance of being in. We can call this list of a probability
    that a data point belonged to the cluster it was chosen to for all data points "confidences". If a given data points
    confidence is in the lowest _blank_ (you set this) percent of confidences, it's marked as an anomaly. This _blank_ is
    from 1 - 100, and is essentially what percent of your data you think are outliers.

    To find the best value for n_clusters, we also have a visualize_elbow_curve method to do that for you. Check the docs
    below if you're interested!

    That's a lot, so let's take a look at its methods!

    NOTE : X SHOULD BE 2D FOR ALL METHODS.

    ____
    Methods

    __init__(n_clusters = 5, retries = 3, max_iters = 200, kmeans_init = True)
        ->> n_clusters is the number of clusters you think are in the data (think of it as the k in KMeans)
        ->> retries is the number of times you want this model to retry. Just like KMeans may need to be tried multiple times,
            we run gaussian mixture models a few times and pick the best solution.
        ->> each retry works an iterative algorithm called Expectation-Maximization (EM). 200 is the maximum amount of iterations
        set by default for the maximum amount of iterations a single retry of EM can take.
        ->> the means of Gaussian Mixture models can be initialized using k-Means (if your data is more than 1D). You can
        set this parameter False if you don't want that.

    fit(X) :
        ->> given X, the model will find the means, standard deviation (or covariance matrices for multivariate data),
        and mixture weights - which are all of the parameters of GMMs
        ->> it will run through the amount of retries specified in the initialization
    predict(X) :
        ->> returns the predicted class for each data point in X
    soft_predict(X) :
        ->> returns the probability for each data point in X that it belonged to each of the n_clusters (defined in the init) clusters
    confidence_samples(X) :
        ->> returns for each data point its probability it belongs to the cluster it has the highest chance of being in.
        This is essentially the confidence of a data point belonging in the data.
    aic() :
        ->> AIC metric that measures how well the data is fit by the Gaussian Mixture
        ->> lower is better
    bic() :
        ->> BIC metric that measures how well the data is fit by the Gaussian Mixture
        ->> lower is better
    anomaly_detect(X, threshold) :
        ->> given this X data, return whether each sample is an anomaly based on whether its "confidence" to the cluster
        it joined is in the bottom threshold percent of all confidences. If its "confidence" is less than the threshold, it's
        marked as an outlier (True) else False. Data which has a low confidence is less likely to truly belong in the data,
        and more likely to be an outlier.
    visualize_clustering(color_dict) :
        ->> only work if your data is in 2 dimensions or more
        ->> visualizes the data and colors each of the data points according to which cluster it joined
        ->> if our color dict was : {0 : "green", 1 : "red", 2 : "orange"}
        ->> here a data point classified to the 0th cluster will be green, a data point to the 1st cluster will be red and
        a data point to the 2nd cluster will be orange.
    visualize_elbow_curve(min_n_clusters = 2, max_n_clusters = 5) :
        ->> to choose the best value of n_clusters, you want to find the elbow on this curve
        ->> this method will try out a bunch of values from n_clusters and then plot its BIC and AIC metrics for each
        value
        ->> the cluster values tried will be from min_n_clusters to max_n_clusters
    """
    def __init__(self, n_clusters = 5, retries = 3, max_iters = 200, kmeans_init = True):
        """
        :param n_clusters : number of clusters assumed in the data
        :param retries : number of times the algorithm will be tried, and then the best solution picked
        :max_iters : maximum number of iterations that the algorithm will be ran for each retry
        :kmeans_init : whether the centroids found by using KMeans should be used to initialize the means in this Gaussian Mixture.
        This will only work if your data is in more than one dimension. If you are using a lot of retries, this may
        not be a good option as it may lead to the same solution over and over again.
        """
        from .cython_mixtures import cy_GaussianMixture
        self.cython_gmm_class = cy_GaussianMixture
        self.cython_gaussian_mixture = cy_GaussianMixture(n_clusters, retries, max_iters, kmeans_init)
        self.k = n_clusters
        self.retries = retries
        self.max_iters = max_iters
        self.kmeans_init = kmeans_init
    def fit(self, X):
        """
        :param X : training data
        this method finds the parameters needed for Gaussian Mixtures to function
        """
        self.X = X
        self.cython_gaussian_mixture.fit(X)
    def predict(self, X):
        """
        :param X : prediction data
        this method returns the cluster each data point in X belongs to
        """
        return self.cython_gaussian_mixture.predict(X)
    def soft_predict(self, X):
        """
        :param X : prediction data
        this method returns the probability each data point belonged to each cluster. This is stored in a matrix
        with the length of X rows and the width of the amount of clusters.
        """
        return self.cython_gaussian_mixture.soft_predict(X)

    def confidence_samples(self, X):
        """
        :param X : prediction data

        This method essentially gives the highest probability in the rows of probabilities in the matrix that is the output
        of the soft_predict() method.

        Translation :

        It's telling you the probability each data point had in the cluster it had
        the largest probability in (and ultimately got assigned to), which is really telling you how confident it is
        that a given data point belongs to the data.
        """
        return self.cython_gaussian_mixture.confidence_samples(X)

    def aic(self):
        """
        Returns the AIC metric (lower is better.)
        """
        return self.cython_gaussian_mixture.aic()

    def bic(self):
        """
        Returns the BIC metric (lower is better.)
        """
        return self.cython_gaussian_mixture.bic()
    def anomaly_detect(self, X, threshold):
        """
        :param X : prediction data
        :param threshold : what percent of the data you believe is an outlier
        Returns whether each data point in X is an anomaly based on whether its confidence in the cluster it was assigned to
        is in the lowest **threshold** percent of all of the confidences.
        """
        return self.cython_gaussian_mixture.anomaly_detect(X, threshold)
    def visualize_clustering(self, color_dict):
        """
        :param color_dict : parameter of the label a cluster was assigned to to its color (must be matplotlib compatible)
        The color dict could be {0 : "green", 1 : "blue", 2 : "red"} for example.
        This method will not work for data that has only 1 dimension (univariate.) It will plot the data you just
        gave in the fit() method.
        """
        is_multivariate = self.cython_gaussian_mixture._is_multivariate()

        if is_multivariate :
            # then visualize
            y_pred = self.cython_gaussian_mixture.predict(self.X)
            import matplotlib.pyplot as plt
            fig = plt.figure()
            for index, prediction in enumerate(y_pred):
                plt.scatter(self.X[index][0], self.X[index][1], color=color_dict[prediction])
            plt.xlabel("x-axis")
            plt.ylabel("y-axis")
            plt.title("Visualized Clustering with Gaussian Mixtures")
            plt.show()

        else :
            raise ValueError("MODEL HAS DATA WITH ONLY 1 DIMENSION. THIS METHOD CANNOT BE USED THEN (VISUALIZATION DIFFICULTIES.)")

    def visualize_elbow_curve(self, min_n_clusters = 2, max_n_clusters = 5):
        """
        :param min_n_clusters : the minimum value of n_clusters to be tried
        :param max_n_clusters : the max value of n_clusters to be tried

        This method tries different values for n_cluster, from min_n_cluster to max_n_cluster, and then plots
        their AIC and BIC metrics. Finding the n_cluster that leads to the "elbow" is probably the optimal n_cluster
        value.
        """
        n_clusters_to_BIC, n_clusters_to_AIC = [], []
        max_n_clusters += 1
        original_k = self.k
        for k in range(min_n_clusters, max_n_clusters) :
            self.k = k
            self.cython_gaussian_mixture = self.cython_gmm_class(k, self.retries, self.max_iters, self.kmeans_init)
            self.cython_gaussian_mixture.fit(self.X)
            n_clusters_to_BIC.append(self.cython_gaussian_mixture.bic())
            n_clusters_to_AIC.append(self.cython_gaussian_mixture.aic())

        self.k = original_k

        self.cython_gaussian_mixture = self.cython_gmm_class(original_k, self.retries, self.max_iters, self.kmeans_init)
        self.cython_gaussian_mixture.fit(self.X)

        import matplotlib.pyplot as plt
        fig = plt.figure()
        plt.plot(np.array(list(range(min_n_clusters, max_n_clusters))), np.array(n_clusters_to_BIC), color = 'green', label = "BIC")
        plt.plot(np.array(list(range(min_n_clusters, max_n_clusters))), np.array(n_clusters_to_AIC), color = "blue", label = "AIC")

        plt.scatter(list(range(min_n_clusters, max_n_clusters)), n_clusters_to_BIC, color = "green")
        plt.scatter(list(range(min_n_clusters, max_n_clusters)), n_clusters_to_AIC, color = "blue")

        plt.xticks([k for k in range(min_n_clusters, max_n_clusters)])
        
        plt.legend()
        plt.xlabel("Number of Clusters")
        plt.ylabel("Information Criteria")
        plt.title("Elbow Curve in Gaussian Mixture Models")
        plt.show()





import pathlib
import sys
from distutils.core import setup
from os import path

import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    read_me_description = fh.read()

with open("requirements.txt") as reqs:
    requirements = reqs.read().split("\n")

non_python_files = [
    "cython_decision_tree_functions.pyx",
    "cython_tsne.pyx",
    "cython_unsupervised_clustering.pyx",
    "cython_naive_bayes.pyx",
    "cython_ensemble_learning.pyx",
    "cython_knn.pyx",
    "cython_mixtures.pyx"
]

version_name = sys.argv[1].replace("refs/tags/", "")
del sys.argv[1]

setup(
    name="sealion",
    packages=setuptools.find_packages(),
    package_data={"": non_python_files},
    include_package_data=True,
    version=version_name,
    license="Apache",
    description="SeaLion is a comprehensive machine learning and data science library for beginners and ml-engineers alike.",
    author="Anish Lakkapragada",
    author_email="anish.lakkapragada@gmail.com",
    url="https://github.com/anish-lakkapragada/SeaLion",
    keywords=["Machine Learning", "Data Science", "Python"],
    install_requires=requirements,
    long_description=read_me_description,
    long_description_content_type="text/markdown",
    python_requires=">=3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
    ],
)
